package  strathclyde.cs.relational.structured;
import strathclyde.cs.relational.operators.*;
import strathclyde.cs.relational.*;	   
import strathclyde.cs.relational.atomic.*;

import java.util.Enumeration;


/**TeX
%
% Wset
%
%\section{ Weighted sets}
We wish to exend our database to be able to represent uncertain
knowledge. It is thus convenient to provide a type of set that
models the axioms of probability theory.
Probability theory defines itself in terms of the following axioms\footnote{
Kolmogorov, "Foundations of Probability", New York, 1950.}:
Let {\em E} be a collection of elements $\alpha \beta \gamma ...$ called elementary
events, and ${\bf F}_E$ a set of subsets of {\em E}; the elements
of the set ${\bf F}_E$ being random events.
\begin{enumerate}
\item {\bf F} is a field of sets.
\item {\bf F} contains {\em E}.
\item $\forall A \in {\bf F}_E, \exists P_E( A) \in {\bf R}$ where $P_E (A)$ is
the probability of $A$.
\item $P_E( E)=1$
\item If $A\times B=\{ \}$ then $$P_E({A+B})=P_E( A) +P_E( B)$$
\end{enumerate}


If we associate with our sets $S$ a 
function $\Pr S A$ with the properties above then we have a model for
probability theory\footnote{ This approach differs from that
of Dey and Sarkar in {\em A Probabalistic Relational Model},
(ACM TODS, Vol 21, No. 3, Sept 1996). In their model the
sum of the probablilities of the tuples in a relation do
not sum to 1.
}.

We need to extend the definitions of the basic set operations
to take probabilities into account.
*/

import strathclyde.cs.relational.*;
import java.util.Random;
public abstract class Wset extends Set
{
 protected double multiplier;
 public Wset(){super();card=0;multiplier=1;}
 /**TeX The variable m is a multiplier used to
 compute the probability of an element from its weight.
 It is a useful scaling factor used to normalise probabilities
 to be equal to 1.*/
 public String toString()  /**TeX
                          Returns a space separted string of elements
                          bracketed thus \{ \} 
						  with each element preceded by its probabilty. */
{ 
    String r="Weighted{";
    Enumeration i;
    
    for(i=members();i.hasMoreElements(); )
    { 	Universal ui=(Universal)i.nextElement()	;
		r=r.concat(" "+ui+":"+P(ui) );
        
    }
    return r.concat("} x "+cardinality());
}

protected abstract double W(Universal b);
/**TeX This returns the current weight of an element in the set.
Weights are proportional to probabilities but do not sum to 1.*/
protected abstract void  addmember(Universal b, double w);
/**TeX  Adds a new member with a given weight,
if a member already exists then the previous weight 
is replaced by the new weight. After executing
{\tt addmember(b,w)} we have $W(b)=w$.
*/
protected abstract void submember(Universal b, double w);
/**TeX Removes a member and part of its weight,
the new weight $W'(b)$ is either zero or
the difference between the old weight and w
 $$W'(b)={\bf max}(W(b)-w,0)$$ */
public double P(Universal u){return multiplier*W(u);}

	
public Universal choice()  

/**TeX
	
If we interpret the basic choice operator $?$ as returning all
members of a set with equal probability then there is an
obvious extension to weighted sets. For large numbers of 
choices, the proportion of times that $(?E)\in A$ will
tend to $\Pr E A$. That is to say, 
the probability function
associated with the elements of the set specifies the probability
that elements will be chosen randomly from the set.
*/

{ 	if (cardinality()==0) return Universal.undefinedValue;
    Random r= new Random();
	while(true)
	{  Enumeration i;
	   Universal e;
	   for(i=members();i.hasMoreElements();){
		   if(r.nextFloat()<P(e=(Universal)i.nextElement()))
			   return e;
	   }
	   
	}
}
public Set join(Doperator comb,Doperator sim, Set b)
/**TeX 

Let $r,s$ be  sets.
Let \begin{equation}\label{join}
p = r(\oplus \join \approx )s
\end{equation}
where \begin{itemize}
        \item$\join$ is the join functional,
        \item $\oplus$ is some dyadic combining operator of type \dyad{Any}{Any}{Any}
        \item $\approx$ is some similarity operator of type 
        \dyad{Any}{Any}{0..1}, two tuples $t,u$ are said to be similar if $(t\approx u)>0$.
\end{itemize}

then we can say that $p$ contains an element corresponding to every pair of elements
in $x,y$ that are similar.
\begin{equation}\label{sim}
\forall i \in r, \forall j \in s|i\approx j, \exists k\in p   
\end{equation}
each such $k= i\oplus j$ is the result of applying the combining operator to
the pairs of similar elements.


Let us define the conditional similarity $C_{ab}$ of two weighted sets $a,b$
to be :
\begin{equation}\label{csim}
 C_{ab}= \sum_{i\in a}\sum_{j \in b}(i \approx j)\times\Pr a i \times \Pr b j 
\end{equation}
This will be a number in the range 0..1.
We can use it to define the probabilities associated with elements of
a joined set.
Thus in  (\ref{join}) and (\ref{sim}) we have
\begin{equation}\label{joinprob}
\Pr{p}{k} = \frac{\sum_{\forall i,j|k=(i\oplus j)} 
(i \approx j)\times\Pr{r}{i} \times \Pr{s}{j}}{C_{rs}}
\end{equation}Note that
\begin{itemize}
        \item
         as it should, the probabilities of the elements 
        sum to unity: $\sum_{a\in p}\Pr p a = 1$,
        \item equation (\ref{joinprob}) generalises equation (\ref{sim}).
\end{itemize}
Taking the definition of conditional similarity $C_{ab}$ given in
(\ref{csim}), the cardinality of joined sets is given by
\begin{equation}
      \card p = C_{rs} \times \card r \times \card s
\end{equation}


*/
{   
	Wset s = (Wset)EmptySet();
	Enumeration i,j;
	double C=Double.MIN_VALUE; // conditional similarity
	for (i=members();i.hasMoreElements();){
		Universal ui = (Universal)i.nextElement();
		double Pi= P(ui);
		for(j=b.members();j.hasMoreElements();){
			Universal uj = (Universal )j.nextElement();
			double Ws,Pj = b.P(uj);
			Number compare = (Number)sim.apply(ui,uj);// similarity
			C+= (Ws=Pj*Pi* compare.doubleValue());
			if (compare.doubleValue()>0){
				Universal nm=comb.apply(ui,uj); // new member
				(s).addmember(nm,s.W(nm)+Ws);
			}
		}	
	}
	// Set multiplier to reciprocal of conditional similarity
	s.multiplier= 1/C;
	// set cardinality of s
	s.card =( cardinality()*b.cardinality())*C;
	return s;
}
public Universal times(Universal b)   
/**TeX
%==================================================================
%\subsection*{Product, a special case of join}
The product of two sets is their intersection.
Intersection of two sets is an operation defined in terms of the
join functional, the equality operator and an operator {\bf left}
that always returns its left argument.
thus for sets $r,s$:
\begin{equation}
r \times s = r ( {\bf left} \join =) s
\end{equation}
Consider  two departments represented as unweighted ( equiprobable ) sets:
\begin{itemize}
\item {\em sales}$=\{$Jo,Mo, Flo,Sue$\}$ with cardinality 4.
\item {\em marketing}$=\{$Lou, Sue, Mo$\}$ with cardinality 3.
\end{itemize}
under the normal interpretation of set theory their intersection would
have cardinality 2.

The conditional similarity of the two sets is given by summing the
product of the probabilities of occurence of Sue and Mo in the two sets 
$=2 \times(\frac 1 4 \times \frac 1 3) = \frac{2}{12}= \frac 1 6$.

The cardinality is then the products of the individual set cardinalities
by the conditional similarity $=\frac{3 \times 4}{6} = 2$ as expected.

Consider now an example using weighted sets.
\begin{itemize}
\item {\em bombings}$=\{$England$_{0.2}$,Ulster$_{0.8}\}$ with cardinality 10, representing
IRA bombings over the last 3 years.
\item {\em meteors}
$=\{$Scotland$_{0.3},$Ulster$_{0.1}$, Wales$_{0.1},$ England$_{0.5}\}$ with cardinality 2
representing probable meteorite falls over the next 3 years.
\end{itemize}
The intersection of the two sets would represent the probabilities that areas
would have an IRA bombing in the last 3 years and a meteorite fall in
the next 3 years.

The conditional similarity is given as $\sum 
\{$England$_{0.2 \times0.5}+$Ulster$_{0.1\times 0.8}\}= 0.18$.

The cardinality is given as $2\times10\times0.18=3.6$.

The resulting set is $$3.6\times
\{ England_{0.55556},Ulster_{0.44444}\}$$
we can interpret this as there being 3.6 expected pairs of bombs and meteorites hitting
the same country.


   */

{   
	if(b instanceof Set)
	{  return join(new Left(),new Equals(),(Set)b); 
	} else return undefinedValue;
}

public Universal  plus(Number weight,Universal b)
{   
	double t,re,c1,c2,r;
	
	r=((Numeric)weight).doubleValue();
	if((r>1)||(r<0) ) return undefinedValue;
	if(b instanceof Set)
	{   
		Set b1=(Set)b;
		Wset e= (Wset)EmptySet();
		Enumeration [] n={members(),b1.members()};
		
		Universal u;
		int i;
		for(i=0;i<2;i++){
			for(;n[i].hasMoreElements();)
			{ 
				Universal ui=(Universal)n[i].nextElement();
				e.addmember(ui,e.P(ui)+(i==0?r*P(ui):(1-r)*b1.P(ui)));				 
			}
		}
		e.multiplier=1;								// unitary scaling of weights
		t=cardinality()+b1.cardinality();	// total cardinality
		re=	cardinality()/t;				// equilibrium partionioning
		c1=	(re/r)*t;						// two estimates of cardinality
		c2= ( (1-re)/(1-r))*t;
		e.card = (c1<c2?c1:c2);             // chose minimum
		return e;
	} else return undefinedValue;
}

/**TeX
%==================================================================
\section*{Union of sets}
\subsection*{Disjoint Union}
Consider the union operation, and more specifically the union operation
on disjoint sets $A,B$. We know by axiom (5) of probability theory
that probabilities of disjoint sets should be additive {\em within the
context of some set $F_E$ to which these constitute subsets}. 
Thus if we let $C=A+B$ we know that 
\begin{equation}\Pr C C =1= \Pr C A + \Pr C B \end{equation}
and that if $A_i < A, B_j < B$ then 
\begin{equation}
 \Pr C {A_i + B_j} = \Pr C {A_i} + \Pr C {B_j}
\end{equation}
However we do not yet have enough information to derive $\Pr C {} $ from
$\Pr A{}$ and $\Pr B{}$.

If we make the assumption that \begin{equation}
\Pr C{A_i}=(\Pr C A) \times(\Pr A {A_i})\end{equation}
then we have an interpretation of the probability function over the
union set in terms of the conditional probability of $A_i$ given that
$A$ occurs. Thus given a real number $r:[0,1)$ as an additional parameter
to + , as in $+_r$, 
we can define a new union operator that has a consistent interpretaion:
 
\begin{equation}A\times B = \{ \}, C= A+_r B, A_i\subset A \Rightarrow
\Pr C {A_i} = r\Pr A {A_i}
\label{plusr}  
\end{equation}

From \ref{plusr} we can deduce that 
\begin{equation}B_j\subset B \Rightarrow
\Pr C {B_j} = (1-r)\Pr B {B_j}
\end{equation}

Clearly it is necessary to have a similar parameterisation of the insertion
operator
\begin{equation} 
A\bullet_r b = A +_r \{b \}
\end{equation}
It is desirable to have an interpretation of the 
 simple unparameterised + operator that retains 
 its original meaning. That is to say, it must preserve the
 meaning of the choice operator. Unparameterised union,
 must result in a set all of whose members are equally
 probable. This can be achieved if
\begin{equation}
a+b=a+_{r_e}b, r_e=\frac{\card{a}}{\card{a}+\card{b}}
\label{equiprob}
\end{equation}
that is to say, the sets are given union probabilities in
proportion to their cardinalities.

\subsection{Cardinality}
The cardinality of the union of two sets is 
normally defined by the equation
\begin{equation}
A\times B=\{\} \Rightarrow \card{(A+B)}= (\card A) + (\card B)
\end{equation}
For weighted sets, this is no longer sufficient.
Instead we want to model situations like the following:
a department has 3 employees, \{ John who is there full time, Robert
and Gordon who each work half  time \}. However, the number of
employees who are there at any one time is 2.
If we randomly select one of the employees present we
will get John 50\% of the time, Robert 25\% and 
Gordon 25\% of the time.

This probability distribution can be obtained by
parameterising the + operator with 0.5.
$$\{John_{0.5},Robert_{0.25},Gordon_{0.25}\}=\{John\}+_{0.5} \{Robert,Gordon\}$$
Is there a treatment of cardinality that is consistent with
our requirements?

If we define $r_e$ as in eqn(\ref{equiprob}), and let $t=(\card A) + (\card B)$,
then we can define the cardinality of weighted unions as follows:
\begin{equation}
A\times B=\{\}, \card{(A+_r B)}= \min (\frac{r_e}{r}t,\frac{1-r_e}{1-r}t)
\end{equation}
This definition of cardinality reduces to the normal
definition in the case of unweighted union.
In the example above we get:$$
\card{(\{John\}+_{0.5} \{Robert,Gordon\})}=\min(\frac{1/3}{1/2}\times 3,\frac{2/3}{1/2}\times 3)=2
$$

\subsection*{Generalised Union}
Having dealt with cardinality and disjoint union we can
now take on the problem of union between weighted sets with
non-null intersections.

Suppose we have two departments {\em sales} =\{John$_{1/3}$,Robert$_{2/3}$\}
and {\em marketing}= \{John$_{1/5}$,Alice$_{2/5},$ Gordon$_{2/5}$\}. {\em Sales} 
set has cardinality 1.5,
{\em marketing} has cardinality 2.5. We can interpret this as John working
half of his time in each department whilst Robert, Alice and Gordon work full-time 
in their respective departments.
The union of the two departments {\em
sales-and-marketing}= {\em sales + marketing}
should have a cardinality 4, as four people
work in it full-time, and each person should have equal weight.
We can generalise eqn(\ref{plusr}) to:

\begin{equation} C= A+_r B, A_i\subset A,A_i\subset B \Rightarrow
\Pr C {A_i} = r\Pr A {A_i}+ (1-r)\Pr B {A_i}
\label{plusgen}  
\end{equation}
In the example of the {\em sales and marketing} department,

{\em sales}+{\em marketing} = {\em sales}$+_{3/8}${\em marketing}

by the normal equiprobable interpretation of union. Johns probability
of being randomly selected from the joint department is then
$\frac 3 8 \times \frac 1 3 + \frac 5 8 \times \frac 1 5 = \frac 1 4 $
as desired.



*/
public Universal plus(Universal b)
{ double t,re;
  if(b instanceof Set){
	  t=cardinality()+((Set)b).cardinality();
	  re = cardinality()/t;
	  return plus( new Numeric(re),b);
  }
  else return undefinedValue;
}
/**TeX
The inherited plus operation is implemented in terms
of the addition of sets using the equiprobable ratios as in eqn(\ref{equiprob}).
 */
}

